from numpy import inf, random
from casadi.tools import entry, struct_symMX
import matplotlib.pyplot as plt
import casadi as ca
from casadi import vertcat, horzcat, DMatrix, SX, MX

# Seed random number generator.
random.seed(927)

# Multiple shooting and integrator parameters
h = 0.25     # MHE sampling time
N_rk = 10     # number of RK4 steps
N_sim = 20    # simulation length

# Plant parameter values
RT = 32.84 # mol atm / L
kp1 = 0.5
km1 = 0.05
kp2 = 0.2
km2 = 0.01

# Standard deviations
sigma_v = 0.25  # Standard deviation of the measurements
sigma_w = 0.001 # Standard deviation for the process noise
sigma_p = 0.5   # Standard deviation for prior

# Make covariance matrices.
P = sigma_p**2 * DMatrix.eye(3) # Covariance for prior.
Q = sigma_w**2 * DMatrix.eye(1)
R = sigma_v**2 * DMatrix.eye(1)

Pinv = sigma_p**-2 * DMatrix.eye(3)
Qinv = sigma_w**-2 * DMatrix.eye(1)
Rinv = sigma_v**-2 * DMatrix.eye(1)

# Initial state estimate
x0_bar = DMatrix([1.0,0.0,4.0])

# Real initial state
x0 = DMatrix([0.5,0.05,0.0])

# States: Concentration of A, B, C
cA = SX.sym('cA')
cB = SX.sym('cB')
cC = SX.sym('cC')
x = vertcat((cA,cB,cC))

# Disturbance on the system dynamics for A, B, C
wA = SX.sym('wA')
wB = SX.sym('wB')
wC = SX.sym('wC')
w = vertcat((wA,wB,wC))

# No parameters
p = SX.sym("p", 0) # vector of length 0

# ODE right hand side derivatives
int1 = kp1*cA - km1*cB*cC
int2 = kp2*cB**2 - km2*cC
cA_dot = -int1
cB_dot =  int1 - 2*int2
cC_dot =  int1 + int2
x_dot = vertcat((cA_dot, cB_dot, cC_dot))

# Continuous-time dynamics
f = ca.SXFunction([x, p], [x_dot])
f.init()

# Discrete time dynamics
F = ca.simpleRK(f, N_rk)

# Measurement function
y = RT*(cA+cB+cC)
F_meas = ca.SXFunction([x,p],[y])
F_meas.init()

# Simulate the system to generate measurements
xk = x0
tk = 0
x_all = [xk]
y_all = []
t_all = [tk]
for k in range(N_sim):
  # Simulate the system and add output disturbance
  [xk_next] = F([xk, 0, h])
  xk_next += sigma_w*random.randn()

  # Get the measurement and add measurement disturbance
  [yk] = F_meas([xk,0])
  yk += sigma_v*random.randn()

  # Advance time
  xk = xk_next
  tk += h

  # Store for plotting
  x_all.append( xk )
  y_all.append( yk )
  t_all.append( tk )

# MHE loop
xMHE = []
for k in range(N_sim):
  N = k+1 # Full information: the horizon keeps increasing!
  print 'Horizon length = %d' % N

  # Multiple Shooting Discretization
  V = struct_symMX([
                    (
                     entry("x",shape=x.shape,repeat=N+1),
                     entry("w",shape=w.shape,repeat=N),
                     entry("v",shape=y.shape,repeat=N),
                     ),
                    ])

  # Cost
  J = 0

  # Penalize the deviation of the initial state from its estimate
  J += 

  # Constraints
  g = []

  for k in range(N):
    # Simulate the system and add output disturbance
    

    # Get the measurement and add measurement disturbance
    

    # Append the constraints to the constraint vector (slide 63)
    g.append(  )
    g.append(  )

    # Penalize state and measurement noise
    J +=  # Penalize state noise
    J +=  # Penalize measurement noise

  # Concatenate constraints
  g = vertcat(g)

  # Formulate subproblem
  nlp = ca.MXFunction(ca.nlpIn(x=V),ca.nlpOut(f=J,g=g))
  nlp.init()

  # Allocate solver
  solver = ca.NlpSolver("ipopt",nlp)
  solver.setOption("print_level",0)
  solver.setOption("print_time",0) 
  solver.init()

  # Solution guess
  v_guess = V(0)
  
  # State bounds


  # Solve the subproblem
  solver.setInput(v_guess,  "x0")
  solver.setInput(v_min,  "lbx")
  solver.setInput(v_max,  "ubx")
  solver.setInput(0,  "lbg") # Equality constraints
  solver.setInput(0,  "ubg")
  solver.evaluate()
  Vopt = V(solver.output('x'))
  
  # Save for plotting
  xMHE.append( Vopt['x',-1] )

# Columns corresponding to different states
t_all = DMatrix(t_all)
x_all = horzcat(x_all).T
xMHE = horzcat(xMHE).T

# Plot the solution
plt.figure(1)
plt.clf()
plt.subplot(1,2,1)
plt.plot(t_all,x_all[:,0],'b')
plt.plot(t_all[:-1],xMHE[:,0],'.b')
plt.plot(t_all,x_all[:,1],'r')
plt.plot(t_all[:-1],xMHE[:,1],'.r')
plt.plot(t_all,x_all[:,2],'k')
plt.plot(t_all[:-1],xMHE[:,2],'.k')
plt.grid(True)

# Plot the output solution
plt.subplot(1,2,2)
plt.plot(t_all[:-1],y_all,'b')
plt.plot(t_all[:-1],RT*(xMHE[:,0]+xMHE[:,1]+xMHE[:,2]),'.b')
plt.show()
